import random
import numpy as np
from torchvision import transforms
from transformers import AutoTokenizer
from ailab.atp_dataset.dataset import AILabDataset
from ailab.atp_finetuner.preprossor import AILabPreprocessor
from ailab.atp_finetuner.build import PreProcessorRg
from ailab.atp_finetuner.constant import Task, Model

@PreProcessorRg.register((Task.text_to_image, Model.stable_diffusion_xl_base_1))
class SDXLPreProcessor(AILabPreprocessor):
    def __init__(self, dataset, preprocessor, model_name):
        super().__init__(dataset, preprocessor)
        self.model_name = model_name
        self.tokenizer_two = None
    
    @classmethod
    def build_preprocessor(cls, model_name:str, dataset: AILabDataset,pc_dir:str,**kwargs):
        pc_name_dir = model_name if pc_dir is None else pc_dir
        tokenizer_one = AutoTokenizer.from_pretrained(
            pc_name_dir,
            subfolder="tokenizer",
            use_fast=False,
        )
        tokenizer_two = AutoTokenizer.from_pretrained(
            pc_name_dir,
            subfolder="tokenizer_2",
            use_fast=False,
        )
        preprocessor_cls =  cls(dataset, tokenizer_one, model_name)
        preprocessor_cls.tokenizer_two = tokenizer_two
        return preprocessor_cls
    
    def process_data(self) ->AILabDataset:
        accelerator = self._accelerator.accelerator_ins
        tokenizer_one =self._preprocessor
        tokenizer_two = self.tokenizer_two
        dataset = self._dataset.to_hf_dataset()
        image_column = "image"
        caption_column = "text"
        resolution = 1024

        def tokenize_prompt(tokenizer, prompt):
            text_inputs = tokenizer(
                prompt,
                padding="max_length",
                max_length=tokenizer.model_max_length,
                truncation=True,
                return_tensors="pt",
            )
            text_input_ids = text_inputs.input_ids
            return text_input_ids
        
        def tokenize_captions(examples, is_train=True):
            captions = []
            for caption in examples[caption_column]:
                if isinstance(caption, str):
                    captions.append(caption)
                elif isinstance(caption, (list, np.ndarray)):
                    # take a random caption if there are multiple
                    captions.append(random.choice(caption) if is_train else caption[0])
                else:
                    raise ValueError(
                        f"Caption column `{caption_column}` should contain either strings or lists of strings."
                    )
            tokens_one = tokenize_prompt(tokenizer_one, captions)
            tokens_two = tokenize_prompt(tokenizer_two, captions)
            return tokens_one, tokens_two
        
        from torchvision import transforms
        train_resize = transforms.Resize(resolution, interpolation=transforms.InterpolationMode.BILINEAR)
        train_crop = transforms.RandomCrop(resolution)
        train_flip = transforms.RandomHorizontalFlip(p=1.0)
        train_transforms = transforms.Compose(
            [
                transforms.ToTensor(),
                transforms.Normalize([0.5], [0.5]),
            ]
        )

        def preprocess_train(examples):
            images = [image.convert("RGB") for image in examples[image_column]]
            # image aug
            original_sizes = []
            all_images = []
            crop_top_lefts = []
            from torchvision.transforms.functional import crop
            for image in images:
                original_sizes.append((image.height, image.width))
                image = train_resize(image)

                y1, x1, h, w = train_crop.get_params(image, (resolution, resolution))
                image = crop(image, y1, x1, h, w)
                crop_top_left = (y1, x1)
                crop_top_lefts.append(crop_top_left)
                image = train_transforms(image)
                all_images.append(image)

            examples["original_sizes"] = original_sizes
            examples["crop_top_lefts"] = crop_top_lefts
            examples["pixel_values"] = all_images
            tokens_one, tokens_two = tokenize_captions(examples)
            examples["input_ids_one"] = tokens_one
            examples["input_ids_two"] = tokens_two
            return examples

        with accelerator.main_process_first():
            train_dataset = dataset["train"].with_transform(preprocess_train, output_all_columns=True)
        return AILabDataset(train_dataset)