#!/usr/bin/env python
# coding:utf-8
"""
@license: Apache License2
@Author: glchen
@time: 2023/03/23
@project: ailab
"""
import json
import os.path
import subprocess
import threading
import traceback

from aiges.core.types import *

try:
    from aiges_embed import ResponseData, Response, DataListNode, DataListCls, callback  # c++
except:
    from aiges.dto import Response, ResponseData, DataListNode, DataListCls, callback

from aiges.sdk import WrapperBase, \
    JsonBodyField, StringBodyField, ImageBodyField, \
    StringParamField
from aiges.utils.log import log, getFileLogger

import io
# from ifly_atp_sdk.huggingface.pipelines import pipeline
import torch

# 使用的模型
model = "runwayml/stable-diffusion-v1-4"
task = "stable-diffusion-text-to-image"
prompt = "a photo of an astronaut riding a horse on mars"


# 定义模型的超参数和输入参数
class UserRequest(object):
    input1 = StringBodyField(key="text", value=prompt.encode("utf-8"))
    input2 = StringParamField(key="task", value=task)


# 定义模型的输出参数
class UserResponse(object):
    accept1 = JsonBodyField(key="result")


# 定义服务推理逻辑
class Wrapper(WrapperBase):
    serviceId = task
    version = "v1"
    requestCls = UserRequest()
    responseCls = UserResponse()
    model = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.filelogger = None
        self.patch_id = {}
        self.wrapper_cls = None
        self.lora_weight_dir = "/home/.atp/lora_weight"
        self.record_dir = "/home/.atp/lora_record"
        # subprocess.check_call("pip install peft==0.6.0 diffusers transformers==4.34.0", shell=True)
        self.lock = threading.Lock()
        self.lora_load()

    def wrapperInit(self, config: {}) -> int:
        base_model_path = os.environ.get("FULL_MODEL_PATH")
        finetune_type = os.environ.get("FINETUNE_TYPE",'lora')
        model_name = os.environ.get("PRETRAINED_MODEL_NAME")
        if not base_model_path:
            log.error("should have environ")
            return -1
        if not os.path.isdir(base_model_path):
            log.error(f"not find the base_model in {base_model_path}")
            return -1
        if finetune_type != 'lora':
            log.error(f"only support lora for diffusion text2image")
            return -1
        from diffusers import StableDiffusionPipeline, DPMSolverMultistepScheduler, StableDiffusionXLPipeline
        from ailab.atp_finetuner.constant import Model
        if model_name in [Model.stable_diffusion_xl_base_1]:
            self.pipeline = StableDiffusionXLPipeline.from_pretrained(
                    base_model_path,
                    torch_dtype=torch.float32,
                )
        else:
            self.pipeline = StableDiffusionPipeline.from_pretrained(
                base_model_path, torch_dtype=torch.float32, use_safetensors=True
            )
        self.pipeline.scheduler = DPMSolverMultistepScheduler.from_config(self.pipeline.scheduler.config)
        self.pipeline.to("cuda")
        self.filelogger = getFileLogger()
        self.filelogger.info(f"wrapperInit success, base model:{base_model_path}, finetune_type:{finetune_type}")
        return 0

    def checkValidLora(self, lora_path):
        weight_file = "pytorch_lora_weights.safetensors"
        if not os.path.isdir(lora_path):
            msg = "not find  %s"%lora_path
            log.error(msg)
            return False, msg
        files = os.listdir(lora_path)
        if not weight_file in files:
            msg = "%s doesnt have file %s" % (lora_path, weight_file)
            log.error(msg)
            return False, msg
        return True, "Check Success..."

    def lora_load(self):
        if not os.path.isdir(self.record_dir):
            log.info("not need load lora now ")
            return
        patch_ids = os.listdir(self.record_dir)
        paths = os.listdir(self.lora_weight_dir)
        for patch_id in patch_ids:
            if not patch_id in paths:
                log.warn("not find this %s"%patch_id)
                continue
            lora_weight_path = os.path.join(self.lora_weight_dir, str(patch_id))
            valid, msg = self.checkValidLora(lora_weight_path)
            if not valid:
                log.warn("not valid lora for this model. %s"%patch_id)
                continue
            self.lock.acquire()
            self.patch_id[patch_id] = lora_weight_path
            self.lock.release()
            log.info(f'success wrapperLoadRes patch_id {patch_id}')

    def wrapperLoadRes(self, reqData: DataListCls, patch_id) -> int:
        if patch_id in self.patch_id:
            log.warn("patch_id has exist.Please first to UnloadRes")
            return 0
        lora_weight_path = os.path.join(self.lora_weight_dir, str(patch_id))

        import io
        import zipfile
        byte_stream = io.BytesIO(reqData.list[0].data)
        # 解压缩 zip 文件到指定目录
        with zipfile.ZipFile(byte_stream, 'r') as zip_ref:
            zip_ref.extractall(lora_weight_path)
        valid, msg = self.checkValidLora(lora_weight_path)
        if not valid:
            return -1

        # write reocrd
        rf = open(os.path.join(self.record_dir, patch_id),"w")
        rf.close()

        log.info("load res patch id:%s, lora weight:%s, result:%s" % (str(patch_id), lora_weight_path, msg))

        self.lock.acquire()
        self.patch_id[patch_id] = lora_weight_path
        self.lock.release()
        return 0

    def wrapperUnloadRes(self, patch_id: int) -> int:
        patch_id = str(patch_id)
        if patch_id not in self.patch_id:
            log.error("patch_id not exist")
            return 0

        self.lock.acquire()
        lora_weight_path = self.patch_id[patch_id]
        del self.patch_id[patch_id]
        self.lock.release()

        if not os.path.exists(lora_weight_path):
            log.error("lora weigth path not exist")
            return 0

        import shutil
        shutil.rmtree(lora_weight_path)

        if os.path.exists(os.path.join(self.record_dir, patch_id)):
            os.remove(os.path.join(self.record_dir, patch_id))
        log.info(f'success wrapperUnloadRes patch_id {patch_id}')

        return 0

    def get_payload_text(self, reqData: DataListCls):
        text = reqData.get('portrait_description').data.decode('utf-8')
        return text

    def _wrapperOnce(self, params: {}, reqData: DataListCls):
        for req in reqData.list:
            self.filelogger.info("got reqdata , key:%s data:%s" % (req.key, req.data))
        self.filelogger.info("got params , %s" % params)
        self.lock.acquire()
        try:
            atp_patch_id = str(params.get('patch_id', '0'))
            if atp_patch_id == 0 or atp_patch_id == '0':
                lora_weight = None
            else:
                lora_weight = self.patch_id[atp_patch_id]
            if lora_weight is not None:
                self.filelogger.info("atp_patch_id:%s, lora:%s" % (str(atp_patch_id), lora_weight))
                self.pipeline.load_lora_weights(lora_weight, adapter_name=str(atp_patch_id))
                self.pipeline.set_adapters(str(atp_patch_id))

            generator = None
            seed = params.get('seed')
            if seed is not None:
                generator = torch.Generator(device="cuda")
                generator = generator.manual_seed(int(seed))
            validation_prompt = self.get_payload_text(reqData)
            w = int(params.get('width', 768))
            h = int(params.get('height', 768))
            negative_prompts = str(params.get('negative_prompts', None))
            image = self.pipeline(validation_prompt, negative_prompt=negative_prompts, num_inference_steps=20, width = w, height = h, generator=generator).images[0]
            self.pipeline.unload_lora_weights()
        except Exception as e:
            stack_trace = traceback.format_exc()
            self.filelogger.error(stack_trace)
        self.lock.release()
        assert image is not None
        # 使用Response封装result
        res = Response()
        resd = ResponseData()
        resd.key = "portrait"
        resd.setDataType(DataImage)
        resd.status = Once
        img_bytes = io.BytesIO()
        image.convert("RGB")
        image.save(img_bytes, format="JPEG")
        resd.setData(img_bytes.getvalue())
        res.list = [resd]
        return res

    def wrapperOnceExec(self, params: {}, reqData: DataListCls, usrTag:str="",persId: int = 0) -> Response:
        self.filelogger.info(f"wrapperOnceExec enter, usrTag:{usrTag}, persId:{persId}")
        res = self._wrapperOnce(params, reqData)
        self.filelogger.info(f'wrapperOnceExec success')
        return res

    def wrapperOnceExecAsync(self, params: {}, reqData: DataListCls, usrTag: str, persId: int = 0):
        self.filelogger.info(f"wrapperOnceExecAsync enter, usrTag:{usrTag}, persId:{persId}")
        res = self._wrapperOnce(params, reqData)
        r = callback(res, usrTag)
        if r != 0:
            self.filelogger.error(f'wrapperOnceExecAsync callback failed, ret code:{r}')
            return r
        self.filelogger.info(f'wrapperOnceExecAsync success')
        return 0

    def wrapperFini(cls) -> int:
        return 0

    def wrapperError(cls, ret: int) -> str:
        if ret == 100:
            return "user error defined here"
        return ""

    '''
        此函数保留测试用，不可删除
    '''

    def wrapperTestFunc(cls, data: [], respData: []):
        pass


if __name__ == '__main__':
    m = Wrapper()
    m.run()
